{$G+,N+}                                                { 286+, 287+ }
{$R-,S-}

{$DEFINE ZBUFFER}

Unit TMappinZ;

Interface

Uses
  Grafix, Vectors, PStack, ZBuffer;

                               (* CONSTANTS *)

{$I CONSTS.PAS}
Const
  TMethod : Word = 1;           { Texture mapping method }
  Zclose = 0.001;               { Z = Zclose - clipping plain equation }
  Max_Edges = 64;               { Max edges available for floors/ceilings }
  ZBConst = 1E9;

                               (* VARIABLES *)

Var
  Scan_Buf_L,
  Scan_Buf_R : Array [0..Screen_Height-1] of LongInt; { Scan buffer for texture mapper }
  ZScan_Buf_L,
  ZScan_Buf_R : Array [0..Screen_Height-1] of Double; { Scan buffer for texture mapper }
  Z_Buf_L,
  Z_Buf_R    : Array [0..Screen_Height-1] of Double;

  Vertices : Array[0..Max_Edges-1] of TVector;  { Current verices for tmap }
  LX, LY : Double;                              { Dimensions of wall to be tmapped }

  XR, YR, ZR : Array[0..Max_Edges] of Double;   { Polygon's coordinates }
  VCNT : Word;                                  { Internal tmapper variable }
  XminI, XmaxI,
  YminI, YmaxI : LongInt;                       { Polygon's screen min & max y coords }


                        (* PROCEDURES AND FUNCTIONS *)

{ Horizontal line texture mapper }
Procedure TLine8 (VOFFS : Word; LLEN : Word; PTEXTURE : Pointer; L3, L1, L2, I3, I1, I2 : Comp);
Procedure TLine16 (VOFFS : Word; LLEN : Word; PTEXTURE : Pointer; L3, L1, L2, I3, I1, I2 : Comp);
{ Polygon edge sanner }
Procedure Scan_Edge (N : Word);
{ Wall texture mapper }
Procedure T_Map (Var Texture : Array of Byte);
{ Portal mapper }
Function P_Map : Boolean;
{ Floor and ceiling texture mapper }
Procedure T_Map_FC (Cnt : Integer; Var Texture : Array of Byte; Height,
                    Xp, Yp : Double; Ang, Yang : Integer);
Procedure Draw_Poly (Cnt : Integer; Color : Byte);


Implementation

{$L TLINE.OBJ}
Procedure TLine8 (VOFFS : Word; LLEN : Word; PTEXTURE : Pointer; L3, L1, L2, I3, I1, I2 : Comp); External;
Procedure TLine16 (VOFFS : Word; LLEN : Word; PTEXTURE : Pointer; L3, L1, L2, I3, I1, I2 : Comp); External;

{$L CEILFLOR.OBJ}
Function Floor (X : Double) : LongInt; Near; External;  { Rounds down }
Function Ceil (X : Double) : LongInt; Near; External;   { Rounds up }

Procedure Scan_Edge (N : Word);
Var
  X, _DX, Z, DZ : Double;
  IsLeftEdge : Boolean;
  X1R, Y1R, X2R, Y2R, Z1R, Z2R : Double;
  Y1I, Y2I, YI, XI : LongInt;
  I, H : Word;
Begin

  { Getting both ends coordinates. }

  X1R := XR[N];
  Y1R := YR[N];
  Z1R := 1/ZR[N];
  X2R := XR[(N+1) mod VCNT];
  Y2R := YR[(N+1) mod VCNT];
  Z2R := 1/ZR[(N+1) mod VCNT];

  { Is line left or right? Determining that. }

  IsLeftEdge := Y2R < Y1R;              { Because of the clockwise order. }

  { Let's perpare beginning values of x and y. }

  If IsLeftEdge then Begin
    Y2I := Ceil(Y2R); Y1I := Ceil(Y1R)-1;
    If Y2I > Y1I then Exit;
    _DX := (X1R - X2R) / (Y1R - Y2R);
    X := X2R + (Y2I - Y2R) * _DX;

    DZ := (Z1R - Z2R) / (Y1R - Y2R);
    Z := Z2R + (Y2I - Y2R) * DZ;

    YI := Y1I; Y1I := Y2I; Y2I := YI
  End
  Else Begin
    Y1I := Ceil(Y1R); Y2I := Ceil(Y2R)-1;
    If Y1I > Y2I then Exit;
    _DX := (X2R - X1R) / (Y2R - Y1R);
    X := X1R + (Y1I - Y1R) * _DX;

    DZ := (Z2R - Z1R) / (Y2R - Y1R);
    Z := Z1R + (Y1I - Y1R) * DZ
  End;

  { Do clipping! }

  If Y1I < 0 then Begin
    X := X - Y1I * _DX;
    Z := Z - Y1I * DZ;
    Y1I := 0
  End;
  If Y2I > Screen_Height-1 then
    Y2I := Screen_Height-1;

  If Y1I > Y2I then Exit;

  { Scan through the line! }
  If IsLeftEdge then
    For YI := Y1I to Y2I do Begin
      ZScan_Buf_L[YI] := X;
      XI := Ceil(X);
      Scan_Buf_L[YI] := XI;
      X := X + _DX;
      Z_Buf_L[YI] := Z;
      Z := Z + DZ
    End
  Else
    For YI := Y1I to Y2I do Begin
      ZScan_Buf_R[YI] := X;
      XI := Ceil(X) - 1;
      Scan_Buf_R[YI] := XI;
      X := X + _DX;
      Z_Buf_R[YI] := Z;
      Z := Z + DZ
    End
End;

Procedure T_Map (Var Texture : Array of Byte);
Var
  Behind : Array [0..3] of Boolean;
  NX, NY : Double;

  P, M, N,
  PM, NP, MN : TVector;
  IndexMin, IndexMax, IndexMinX, IndexMaxX, I, K : Integer;
  Y, XLI, XRI : LongInt;
  YminR, YmaxR,
  D, DDX, DDY,
  DU, DUDX, DUDY,
  DV, DVDX, DVDY,
  WD, WDU, WDV,
  IZ : Double;
  X, U, V : LongInt;
  C : Byte;
  PP : PChar;
  Tmp1, Tmp2 : Integer;

  Z, DZ : Double;
  ZI, DZI : LongInt;
  ZBN, ZBO,
  ZLN : Word;
  PZB : PZBuf50;
Label
  LSkip;
Begin
{
  For I := 0 to 3 do Begin
    Vertices[I].x := Vertices[I].x + 0.5;
    Vertices[I].y := Vertices[I].y + 0.5
  End;
}
  For I := 0 to 3 do
    Behind[I] := Vertices[I].Z < Zclose;

  { Let's find the corner points }

  VCNT := 0;
  For I := 0 to 3 do Begin
    If not Behind[I] then Begin
      XR[VCNT] := Vertices[I].X / Vertices[I].Z;
      YR[VCNT] := Vertices[I].Y / Vertices[I].Z;
      ZR[VCNT] := Vertices[I].Z;
      Inc (VCNT);
    End;
    K := (I+1) mod 4;
    If Behind[I] <> Behind[K] then Begin
      IZ := (Zclose-Vertices[I].Z) / (Vertices[K].Z-Vertices[I].Z);
      NX := Vertices[I].X + (Vertices[K].X-Vertices[I].X) * IZ;
      NY := Vertices[I].Y + (Vertices[K].Y-Vertices[I].Y) * IZ;
      XR[VCNT] := NX / Zclose;
      YR[VCNT] := NY / Zclose;
      ZR[VCNT] := ZClose;
      Inc (VCNT)
    End
  End;
  If VCNT < 3 then Exit;

  { Let's transform the points to the screen coordinate system. }

  For I := 0 to VCNT-1 do Begin
    XR[I] := Center_X+256*XR[I];
    YR[I] := Center_Y+256*YR[I]
  End;

  { Let's find the top and the bottom points. }

  IndexMin := 0;
  IndexMax := 0;
  IndexMinX := 0;
  IndexMaxX := 0;
  For I := 0 to VCNT-1 do Begin
    If YR[I] < YR[IndexMin] then IndexMin := I;
    If YR[I] > YR[IndexMax] then IndexMax := I;
    If XR[I] < XR[IndexMinX] then IndexMinX := I;
    If XR[I] > XR[IndexMaxX] then IndexMaxX := I
  End;

  YminR := YR[IndexMin];
  YmaxR := YR[IndexMax];
  YminI := Ceil(YminR);
  YmaxI := Ceil(YmaxR)-1;

  XminI := Ceil(XR[IndexMinX]);
  XmaxI := Ceil(XR[IndexMaxX])-1;

  { Is the polygon visible at the screen? Skip if it isn't. }

  If (YminI > YmaxI) or (XminI > XmaxI) then Exit;
  If (YminI > PBoundaries^[SSP-1].Ymax) or
     (YmaxI < PBoundaries^[SSP-1].Ymin) or
     (XminI > PBoundaries^[SSP-1].Xmax) or
     (XmaxI < PBoundaries^[SSP-1].Xmin) then Exit;

  Tmp1 := PBoundaries^[SSP-1].Ymin;
  If YminI < Tmp1 then YminI := Tmp1;
  Tmp1 := PBoundaries^[SSP-1].Ymax;
  If YmaxI > Tmp1 then YmaxI := Tmp1;

  { Do scan for each edge. }

  For I := 0 to VCNT-1 do
    Scan_Edge (I);

  { And now we're preparing some vectors and constants }
  { to use in the inner loops. }

  P := Vertices[0];
  Vector_Sub (Vertices[1], Vertices[0], M);
  Vector_Sub (Vertices[3], Vertices[0], N);
  Cross_Product (M, N, MN);
  Cross_Product (P, M, PM);
  Cross_Product (N, P, NP);

  DDX := MN.X;
  DDY := MN.Y;
  D := MN.Z*256 + (YminI-Center_Y)*DDY;

  DUDX := NP.X;
  DUDY := NP.Y;
  DU := NP.Z*256 + (YminI-Center_Y)*DUDY;

  DVDX := PM.X;
  DVDY := PM.Y;
  DV := PM.Z*256 + (YminI-Center_Y)*DVDY;

  { Going down through the polygon screen projection. }
{
  ZBN := YminI div 50;
  ZBO := YminI mod 50;
  ZLN := 50 - ZBO;
}
  Asm
    Mov   AX, Word Ptr YminI
    Xor   DX, DX
    Mov   BX, 50
    Div   BX
    Mov   ZBN, AX
    Mov   ZBO, DX
    Sub   DX, 50
    Neg   DX
    Mov   ZLN, DX
  End;
  PZB := PZBuf[ZBN]; Inc (Word(PZB), (ZBO shl 10) + (ZBO shl 8));

  For Y := YminI to YmaxI do Begin

    { Getting left and right boundaries XLI and XRI. }

    XLI := Scan_Buf_L[Y];
    XRI := Scan_Buf_R[Y];

    If XLI > XRI then Goto LSkip;
    Tmp1 := PBoundaries^[SSP-1].Boundary[Y,0];
    Tmp2 := PBoundaries^[SSP-1].Boundary[Y,1];
    If Tmp1 > Tmp2 then Goto LSkip;
    If (XLI > Tmp2) or
       (XRI < Tmp1) then Goto LSkip;

    Z := Z_Buf_L[Y];
    If XRI <> XLI then
      DZ := (Z_Buf_R[Y] - Z) / (ZScan_Buf_R[Y] - ZScan_Buf_L[Y])
    Else DZ := 0;

    If XLI < Tmp1 then Begin
      Z := Z + (Tmp1 - ZScan_Buf_L[Y]) * DZ;
      XLI := Tmp1
    End;
    If XRI > Tmp2 then
      XRI := Tmp2;

    WD := D + (XLI-Center_X)*DDX;
    WDU := DU + (XLI-Center_X)*DUDX;
    WDV := DV + (XLI-Center_X)*DVDX;

    { Go ahead! Texturate the horizontal line! }

    PP := Ptr (VSegA, XLI);
    Inc (PP, (Y shl 6)+(Y shl 8));

    Case TMethod of
      0: { Pascal source + floating point. } Begin
          For X := XLI to XRI do Begin
            IZ := 1 / WD;
            U := Trunc (WDU * LX * IZ);
            V := Trunc (WDV * LY * IZ);
            C := Texture[((V and 127) shl Texture_Bits) + (U and 127)];
            PP^ := Char(C);                     { PutPixel (X, Y, C); }
            Inc (PP);
            WD := WD + DDX;
            WDU := WDU + DUDX;
            WDV := WDV + DVDX
          End
        End;
      1: { Asm source + mixed fixed and floating point + }
         { scan line subdivision for every 8 pixels. }
        TLine8 (Ofs(PP^), XRI-XLI+1, Addr(Texture),
               WD, WDU*LX, WDV*LY, DDX, DUDX*LX, DVDX*LY);
      2: { Asm source + mixed fixed and floating point + }
         { scan line subdivision for every 16 pixels. }
        TLine16 (Ofs(PP^), XRI-XLI+1, Addr(Texture),
               WD, WDU*LX, WDV*LY, DDX, DUDX*LX, DVDX*LY)
    End;

{$IFDEF ZBUFFER}
    ZI := Trunc(Z*ZBConst);
    DZI := Trunc(DZ*ZBConst);
    Asm
      LES   DI, PZB
      Mov   AX, Word Ptr XLI
      Mov   CX, Word Ptr XRI
      Sub   CX, AX
      Inc   CX
      Shl   AX, 2
      Add   DI, AX
      db $66; Mov AX, Word Ptr ZI
      db $66; Mov BX, Word Ptr DZI
@@zcyc:
      db $66; Mov ES:[DI], AX
      db $66; Add AX, BX
      Add   DI, 4
      Loop  @@zcyc
    End;
{$ENDIF}

LSkip:
    D := D + DDY;
    DU := DU + DUDY;
    DV := DV + DVDY;
    Inc (ZBO);
    Dec (ZLN);
    If ZLN = 0 then Begin
      ZBO := 0;
      ZLN := 50;
      Inc (ZBN);
      If ZBN < 4 then
        PZB := PZBuf[ZBN]
    End
    Else
      Inc (Word(PZB), 320*4)
  End
End;

Function P_Map : Boolean;
Var
  Behind : Array [0..3] of Boolean;
  NX, NY : Double;

  IndexMin, IndexMax, IndexMinX, IndexMaxX, I, K : Integer;
  Y, XLI, XRI : LongInt;
  YminR, YmaxR,
  IZ : Double;

  R : Boolean;
  Dr : Boolean;
  Tmp1, Tmp2 : Integer;
Label
  LSkip, LEnd;
Begin
  R := False;
{
  For I := 0 to 3 do Begin
    Vertices[I].x := Vertices[I].x + 0.5;
    Vertices[I].y := Vertices[I].y + 0.5
  End;
}
  For I := 0 to 3 do
    Behind[I] := Vertices[I].Z < Zclose;

  { Let's find the corner points }

  VCNT := 0;
  For I := 0 to 3 do Begin
    If not Behind[I] then Begin
      XR[VCNT] := Vertices[I].X / Vertices[I].Z;
      YR[VCNT] := Vertices[I].Y / Vertices[I].Z;
      Inc (VCNT);
    End;
    K := (I+1) mod 4;
    If Behind[I] <> Behind[K] then Begin
      IZ := (Zclose-Vertices[I].Z) / (Vertices[K].Z-Vertices[I].Z);
      NX := Vertices[I].X + (Vertices[K].X-Vertices[I].X) * IZ;
      NY := Vertices[I].Y + (Vertices[K].Y-Vertices[I].Y) * IZ;
      XR[VCNT] := NX / Zclose;
      YR[VCNT] := NY / Zclose;
      Inc (VCNT)
    End
  End;
  If VCNT < 3 then Goto LEnd;

  { Let's transform the points to the screen coordinate system. }

  For I := 0 to VCNT-1 do Begin
    XR[I] := Center_X+256*XR[I];
    YR[I] := Center_Y+256*YR[I]
  End;

  { Let's find the top and the bottom points. }

  IndexMin := 0;
  IndexMax := 0;
  IndexMinX := 0;
  IndexMaxX := 0;
  For I := 0 to VCNT-1 do Begin
    If YR[I] < YR[IndexMin] then IndexMin := I;
    If YR[I] > YR[IndexMax] then IndexMax := I;
    If XR[I] < XR[IndexMinX] then IndexMinX := I;
    If XR[I] > XR[IndexMaxX] then IndexMaxX := I
  End;

  YminR := YR[IndexMin];
  YmaxR := YR[IndexMax];
  YminI := Ceil(YminR);
  YmaxI := Ceil(YmaxR)-1;

  XminI := Ceil(XR[IndexMinX]);
  XmaxI := Ceil(XR[IndexMaxX])-1;

  { Is the polygon visible at the screen? Skip if it isn't. }

  If (YminI > YmaxI) or (XminI > XmaxI) then Goto LEnd;
  If (YminI > PBoundaries^[SSP-2].Ymax) or
     (YmaxI < PBoundaries^[SSP-2].Ymin) or
     (XminI > PBoundaries^[SSP-2].Xmax) or
     (XmaxI < PBoundaries^[SSP-2].Xmin) then Goto LEnd;

  Tmp1 := PBoundaries^[SSP-2].Ymin;
  If YminI < Tmp1 then YminI := Tmp1;
  Tmp1 := PBoundaries^[SSP-2].Ymax;
  If YmaxI > Tmp1 then YmaxI := Tmp1;

  { Do scan for each edge. }

  For I := 0 to VCNT-1 do
    Scan_Edge (I);

  PBoundaries^[SSP-1].Ymin := YminI;
  PBoundaries^[SSP-1].Ymax := YmaxI;
  XminI := Screen_Width;
  XmaxI := -1;
  Dr := False;

  For I := YminI to YmaxI do Begin
    XLI := Scan_Buf_L[I];
    XRI := Scan_Buf_R[I];
    If XLI > XRI then Goto LSkip;

    Tmp1 := PBoundaries^[SSP-2].Boundary[I,0];
    Tmp2 := PBoundaries^[SSP-2].Boundary[I,1];
    If Tmp1 > Tmp2 then Goto LSkip;

    If (XLI > Tmp2) or (XRI < Tmp1) then Begin
LSkip:
{      If not Dr then Inc (PBoundaries^[SSP-1].Ymin)
      Else Begin}
        PBoundaries^[SSP-1].Boundary[I,0] := Screen_Width;
        PBoundaries^[SSP-1].Boundary[I,1] := -1;
{      End;}
      Continue
    End;

    Dr := True;

    If XLI < Tmp1 then
      XLI := Tmp1;
    If XRI > Tmp2 then
      XRI := Tmp2;
    PBoundaries^[SSP-1].Boundary[I,0] := XLI;
    PBoundaries^[SSP-1].Boundary[I,1] := XRI;

    If XLI < XminI then XminI := XLI;
    If XRI > XmaxI then XmaxI := XRI
  End;

  If (not Dr) or (PBoundaries^[SSP-1].Ymin > PBoundaries^[SSP-1].Ymax) or
     (XminI > XmaxI) then Goto LEnd;

  PBoundaries^[SSP-1].Xmin := XminI;
  PBoundaries^[SSP-1].Xmax := XmaxI;

  R := True;
LEnd:
  P_Map := R;
End;

Procedure T_Map_FC (Cnt : Integer; Var Texture : Array of Byte; Height,
                    Xp, Yp : Double; Ang, Yang : Integer);
Var
  Behind : Array [0..Max_Edges] of Boolean;
  NX, NY : Double;

  P, M, N,
  PM, NP, MN : TVector;
  IndexMin, IndexMax, IndexMinX, IndexMaxX, I, K : Integer;
  Y, XLI, XRI : LongInt;
  YminR, YmaxR,
  D, DDX, DDY,
  DU, DUDX, DUDY,
  DV, DVDX, DVDY,
  WD, WDU, WDV,
  IZ : Double;
  X, U, V : LongInt;
  C : Byte;
  PP : PChar;
  Tmp1, Tmp2 : Integer;
  x1, y1, x1s, z1s : Double;

  Z, DZ : Double;
  ZI, DZI : LongInt;
  ZBN, ZBO,
  ZLN : Word;
  PZB : PZBuf50;
Label
  LSkip;
Begin
{
  For I := 0 to Cnt-1 do Begin
    Vertices[I].x := Vertices[I].x + 0.5;
    Vertices[I].y := Vertices[I].y + 0.5
  End;
}
  For I := 0 to Cnt-1 do
    Behind[I] := Vertices[I].Z < Zclose;

  { Let's find the corner points }

  VCNT := 0;
  For I := 0 to Cnt-1 do Begin
    If not Behind[I] then Begin
      XR[VCNT] := Vertices[I].X / Vertices[I].Z;
      YR[VCNT] := Vertices[I].Y / Vertices[I].Z;
      ZR[VCNT] := Vertices[I].Z;
      Inc (VCNT);
    End;
    K := (I+1) mod Cnt;
    If Behind[I] <> Behind[K] then Begin
      IZ := (Zclose-Vertices[I].Z) / (Vertices[K].Z-Vertices[I].Z);
      NX := Vertices[I].X + (Vertices[K].X-Vertices[I].X) * IZ;
      NY := Vertices[I].Y + (Vertices[K].Y-Vertices[I].Y) * IZ;
      XR[VCNT] := NX / Zclose;
      YR[VCNT] := NY / Zclose;
      ZR[VCNT] := ZClose;
      Inc (VCNT)
    End
  End;
  If VCNT < 3 then Exit;

  { Let's transform the points to the screen coordinate system. }

  For I := 0 to VCNT-1 do Begin
    XR[I] := Center_X+256*XR[I];
    YR[I] := Center_Y+256*YR[I]
  End;

  { Let's find the top and the bottom points. }

  IndexMin := 0;
  IndexMax := 0;
  IndexMinX := 0;
  IndexMaxX := 0;
  For I := 0 to VCNT-1 do Begin
    If YR[I] < YR[IndexMin] then IndexMin := I;
    If YR[I] > YR[IndexMax] then IndexMax := I;
    If XR[I] < XR[IndexMinX] then IndexMinX := I;
    If XR[I] > XR[IndexMaxX] then IndexMaxX := I
  End;

  YminR := YR[IndexMin];
  YmaxR := YR[IndexMax];
  YminI := Ceil(YminR);
  YmaxI := Ceil(YmaxR)-1;

  XminI := Ceil(XR[IndexMinX]);
  XmaxI := Ceil(XR[IndexMaxX])-1;

  { Is the polygon visible at the screen? Skip if it isn't. }

  If (YminI > YmaxI) or (XminI > XmaxI) then Exit;
  If (YminI > PBoundaries^[SSP-1].Ymax) or
     (YmaxI < PBoundaries^[SSP-1].Ymin) or
     (XminI > PBoundaries^[SSP-1].Xmax) or
     (XmaxI < PBoundaries^[SSP-1].Xmin) then Exit;

  Tmp1 := PBoundaries^[SSP-1].Ymin;
  If YminI < Tmp1 then YminI := Tmp1;
  Tmp1 := PBoundaries^[SSP-1].Ymax;
  If YmaxI > Tmp1 then YmaxI := Tmp1;

  { Do scan for each edge. }

  For I := 0 to VCNT-1 do
    Scan_Edge (I);

  Ang := Ang-90;
  If Ang < 0 then Inc (Ang, 360);

  x1 := -Xp; y1 := -Yp;
  x1s := y1*SinTab[Ang] + x1*CosTab[Ang];
  z1s := y1*CosTab[Ang] - x1*SinTab[Ang];
  P.x := x1s;
  P.y := Height*CosTab[Yang] - z1s*SinTab[Yang];
  P.z := Height*SinTab[Yang] + z1s*CosTab[Yang];

  x1 := Texture_Size-Xp; y1 := -Yp;
  x1s := y1*SinTab[Ang] + x1*CosTab[Ang];
  z1s := y1*CosTab[Ang] - x1*SinTab[Ang];
  N.x := x1s - P.x;
  N.y := Height*CosTab[Yang] - z1s*SinTab[Yang] - P.y;
  N.z := Height*SinTab[Yang] + z1s*CosTab[Yang] - P.z;

  x1 := -Xp; y1 := Texture_Size-Yp;
  x1s := y1*SinTab[Ang] + x1*CosTab[Ang];
  z1s := y1*CosTab[Ang] - x1*SinTab[Ang];
  M.x := x1s - P.x;
  M.y := Height*CosTab[Yang] - z1s*SinTab[Yang] - P.y;
  M.z := Height*SinTab[Yang] + z1s*CosTab[Yang] - P.z;

  Cross_Product (M, N, MN);
  Cross_Product (P, M, PM);
  Cross_Product (N, P, NP);

  DDX := MN.X;
  DDY := MN.Y;
  D := MN.Z*256 + (YminI-Center_Y)*DDY;

  DUDX := NP.X;
  DUDY := NP.Y;
  DU := NP.Z*256 + (YminI-Center_Y)*DUDY;

  DVDX := PM.X;
  DVDY := PM.Y;
  DV := PM.Z*256 + (YminI-Center_Y)*DVDY;

{
  ZBN := YminI div 50;
  ZBO := YminI mod 50;
  ZLN := 50 - ZBO;
}
  Asm
    Mov   AX, Word Ptr YminI
    Xor   DX, DX
    Mov   BX, 50
    Div   BX
    Mov   ZBN, AX
    Mov   ZBO, DX
    Sub   DX, 50
    Neg   DX
    Mov   ZLN, DX
  End;
  PZB := PZBuf[ZBN]; Inc (Word(PZB), (ZBO shl 10) + (ZBO shl 8));

  For Y := YminI to YmaxI do Begin

    { Getting left and right boundaries XLI and XRI. }

    XLI := Scan_Buf_L[Y];
    XRI := Scan_Buf_R[Y];

    If XLI > XRI then Goto LSkip;
    Tmp1 := PBoundaries^[SSP-1].Boundary[Y,0];
    Tmp2 := PBoundaries^[SSP-1].Boundary[Y,1];
    If Tmp1 > Tmp2 then Goto LSkip;
    If (XLI > Tmp2) or
       (XRI < Tmp1) then Goto LSkip;

    Z := Z_Buf_L[Y];
    If XRI <> XLI then
      DZ := (Z_Buf_R[Y] - Z) / (ZScan_Buf_R[Y] - ZScan_Buf_L[Y])
    Else DZ := 0;

    If XLI < Tmp1 then Begin
      Z := Z + (Tmp1 - ZScan_Buf_L[Y]) * DZ;
      XLI := Tmp1
    End;
    If XRI > Tmp2 then
      XRI := Tmp2;

    WD := D + (XLI-Center_X)*DDX;
    WDU := DU + (XLI-Center_X)*DUDX;
    WDV := DV + (XLI-Center_X)*DVDX;

    { Go ahead! Texturate the horizontal line! }

    PP := Ptr (VSegA, XLI);
    Inc (PP, (Y shl 6)+(Y shl 8));

    Case TMethod of
      0: { Pascal source + floating point. } Begin
          For X := XLI to XRI do Begin
            IZ := Texture_Size / WD;
            U := Trunc (WDU * IZ);
            V := Trunc (WDV * IZ);
            C := Texture[((V and 127) shl Texture_Bits) + (U and 127)];
            PP^ := Char(C);                     { PutPixel (X, Y, C); }
            Inc (PP);
            WD := WD + DDX;
            WDU := WDU + DUDX;
            WDV := WDV + DVDX
          End
        End;
      1: { Asm source + mixed fixed and floating point + }
         { scan line subdivision for every 8 pixels. }
        TLine8 (Ofs(PP^), XRI-XLI+1, Addr(Texture),
               WD, WDU*Texture_Size, WDV*Texture_Size,
               DDX, DUDX*Texture_Size, DVDX*Texture_Size);
      2: { Asm source + mixed fixed and floating point + }
         { scan line subdivision for every 16 pixels. }
        TLine16 (Ofs(PP^), XRI-XLI+1, Addr(Texture),
               WD, WDU*Texture_Size, WDV*Texture_Size,
               DDX, DUDX*Texture_Size, DVDX*Texture_Size)
    End;

{$IFDEF ZBUFFER}
    ZI := Trunc(Z*ZBConst);
    DZI := Trunc(DZ*ZBConst);
    Asm
      LES   DI, PZB
      Mov   AX, Word Ptr XLI
      Mov   CX, Word Ptr XRI
      Sub   CX, AX
      Inc   CX
      Shl   AX, 2
      Add   DI, AX
      db $66; Mov AX, Word Ptr ZI
      db $66; Mov BX, Word Ptr DZI
@@zcyc:
      db $66; Mov ES:[DI], AX
      db $66; Add AX, BX
      Add   DI, 4
      Loop  @@zcyc
    End;
{$ENDIF}

LSkip:
    D := D + DDY;
    DU := DU + DUDY;
    DV := DV + DVDY;
    Inc (ZBO);
    Dec (ZLN);
    If ZLN = 0 then Begin
      ZBO := 0;
      ZLN := 50;
      Inc (ZBN);
      If ZBN < 4 then
        PZB := PZBuf[ZBN]
    End
    Else
      Inc (Word(PZB), 320*4)
  End
End;

Procedure Draw_Poly (Cnt : Integer; Color : Byte);
Var
  Behind : Array [0..3] of Boolean;
  NX, NY : Double;

  IndexMin, IndexMax, IndexMinX, IndexMaxX, I, K : Integer;
  Y, XLI, XRI : LongInt;
  YminR, YmaxR, IZ : Double;
  X, U, V : LongInt;
  C : Byte;
  PP : PChar;
  Tmp1, Tmp2 : Integer;

  Z, DZ : Double;
  ZI, DZI : LongInt;
  ZBN, ZBO,
  ZLN : Word;
  PZB : PZBuf50;
Label
  LSkip;
Begin
{
  For I := 0 to Cnt-1 do Begin
    Vertices[I].x := Vertices[I].x + 0.5;
    Vertices[I].y := Vertices[I].y + 0.5
  End;
}
  For I := 0 to Cnt-1 do
    Behind[I] := Vertices[I].Z < Zclose;

  { Let's find the coner points }

  VCNT := 0;
  For I := 0 to Cnt-1 do Begin
    If not Behind[I] then Begin
      XR[VCNT] := Vertices[I].X / Vertices[I].Z;
      YR[VCNT] := Vertices[I].Y / Vertices[I].Z;
      ZR[VCNT] := Vertices[I].Z;
      Inc (VCNT);
    End;
    K := (I+1) mod Cnt;
    If Behind[I] <> Behind[K] then Begin
      IZ := (Zclose-Vertices[I].Z) / (Vertices[K].Z-Vertices[I].Z);
      NX := Vertices[I].X + (Vertices[K].X-Vertices[I].X) * IZ;
      NY := Vertices[I].Y + (Vertices[K].Y-Vertices[I].Y) * IZ;
      XR[VCNT] := NX / Zclose;
      YR[VCNT] := NY / Zclose;
      ZR[VCNT] := ZClose;
      Inc (VCNT)
    End
  End;
  If VCNT < 3 then Exit;

  { Let's transform the points to the screen coordinate system. }

  For I := 0 to VCNT-1 do Begin
    XR[I] := Center_X+256*XR[I];
    YR[I] := Center_Y+256*YR[I]
  End;

  { Let's find the top and the bottom points. }

  IndexMin := 0;
  IndexMax := 0;
  IndexMinX := 0;
  IndexMaxX := 0;
  For I := 0 to VCNT-1 do Begin
    If YR[I] < YR[IndexMin] then IndexMin := I;
    If YR[I] > YR[IndexMax] then IndexMax := I;
    If XR[I] < XR[IndexMinX] then IndexMinX := I;
    If XR[I] > XR[IndexMaxX] then IndexMaxX := I
  End;

  YminR := YR[IndexMin];
  YmaxR := YR[IndexMax];
  YminI := Ceil(YminR);
  YmaxI := Ceil(YmaxR)-1;

  XminI := Ceil(XR[IndexMinX]);
  XmaxI := Ceil(XR[IndexMaxX])-1;

  { Is the polygon visible at the screen? Skip if it isn't. }

  If (YminI > YmaxI) or (XminI > XmaxI) then Exit;
  If (YminI > PBoundaries^[SSP-1].Ymax) or
     (YmaxI < PBoundaries^[SSP-1].Ymin) or
     (XminI > PBoundaries^[SSP-1].Xmax) or
     (XmaxI < PBoundaries^[SSP-1].Xmin) then Exit;

  Tmp1 := PBoundaries^[SSP-1].Ymin;
  If YminI < Tmp1 then YminI := Tmp1;
  Tmp1 := PBoundaries^[SSP-1].Ymax;
  If YmaxI > Tmp1 then YmaxI := Tmp1;

  { Do scan for each edge. }

  For I := 0 to VCNT-1 do
    Scan_Edge (I);

  { Going down through the polygon screen projection. }
{
  ZBN := YminI div 50;
  ZBO := YminI mod 50;
  ZLN := 50 - ZBO;
}
  Asm
    Mov   AX, Word Ptr YminI
    Xor   DX, DX
    Mov   BX, 50
    Div   BX
    Mov   ZBN, AX
    Mov   ZBO, DX
    Sub   DX, 50
    Neg   DX
    Mov   ZLN, DX
  End;
  PZB := PZBuf[ZBN]; Inc (Word(PZB), (ZBO shl 10) + (ZBO shl 8));

  For Y := YminI to YmaxI do Begin

    { Getting left and right boundaries XLI and XRI. }

    XLI := Scan_Buf_L[Y];
    XRI := Scan_Buf_R[Y];

    If XLI > XRI then Goto LSkip;
    Tmp1 := PBoundaries^[SSP-1].Boundary[Y,0];
    Tmp2 := PBoundaries^[SSP-1].Boundary[Y,1];
    If Tmp1 > Tmp2 then Goto LSkip;
    If (XLI > Tmp2) or
       (XRI < Tmp1) then Goto LSkip;

    Z := Z_Buf_L[Y];
    If XRI <> XLI then
      DZ := (Z_Buf_R[Y] - Z) / (ZScan_Buf_R[Y] - ZScan_Buf_L[Y])
    Else DZ := 0;

    If XLI < Tmp1 then Begin
      Z := Z + (Tmp1 - ZScan_Buf_L[Y]) * DZ;
      XLI := Tmp1
    End;
    If XRI > Tmp2 then
      XRI := Tmp2;

{$IFDEF ZBUFFER}
    ZI := Trunc(Z*ZBConst);
    DZI := Trunc(DZ*ZBConst);
    Asm
      Push  DS
      LES   DI, PZB
      Mov   AX, Word Ptr XLI
      Mov   CX, Word Ptr XRI
      Sub   CX, AX
      Inc   CX                          { CX = counter }
      Shl   AX, 2
      Add   DI, AX                      { ES:DI = ZBuffer address }

      Mov   DS, VSegA
      Mov   AX, Word Ptr Y
      Xchg  AH, AL
      Mov   SI, AX
      Shr   AX, 2
      Add   SI, AX
      Add   SI, Word Ptr XLI            { DS:SI = screen address }
      Mov   AL, Color                   { BL = desired color }

      db $66; Mov DX, Word Ptr ZI
      db $66; Mov BX, Word Ptr DZI
@@zcyc:
      db $66; Cmp DX, ES:[DI]
      JC    @@skip_pixel
      db $66; Mov ES:[DI], DX
      Mov   DS:[SI], AL
@@skip_pixel:
      db $66; Add DX, BX
      Add   DI, 4
      Inc   SI
      Loop  @@zcyc
      Pop   DS
    End;
{$ENDIF}

LSkip:
    Inc (ZBO);
    Dec (ZLN);
    If ZLN = 0 then Begin
      ZBO := 0;
      ZLN := 50;
      Inc (ZBN);
      If ZBN < 4 then
        PZB := PZBuf[ZBN]
    End
    Else
      Inc (Word(PZB), 320*4)
  End
End;

End.
